--------------------------------------------------------------------------------
--                    Digilent DIO2 board controller                          --
--                   Designed for Xilinx Spartan IIE                          --
--          Author: Milos Becvar, CVUT Prague, becvarm@fel.cvut.cz            --
--                             based on solution by                           -- 
--                  Jiri Halak, CVUT Prague, halakj1@fel.cvut.cz              --
--                         Source: VHDL code                                  --
--                   Version: 1.0 (Working prototype)	                      --
--          Please send me any error descriptions you find out.               --
--------------------------------------------------------------------------------

LIBRARY IEEE;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;

ENTITY dio2_interface IS
  PORT( clk : IN std_logic;                       -- clock signal
       reset: IN std_logic;                       -- global async. reset 
       -- dio2 external signals
 cs, we, oe : OUT std_logic;                      -- CPLD drive signals
       addr : OUT std_logic_vector(6 DOWNTO 0);   -- CPLD device address
       data : INOUT std_logic_vector(7 DOWNTO 0); -- DIO2 data bus
    dio2clk : OUT std_logic;                      -- DIO2 peripherals clock signal
     lcd_rw : OUT std_logic;                      -- LCD read/write
     lcd_rs : OUT std_logic;                      -- LCD function select
      lcd_e : OUT std_logic;                      -- LCD enable
      -- internal peripheral interface
  clocks    : OUT std_logic_vector(24 downto 0);
  res       : OUT std_logic; -- internal reset
  sseg      : IN  std_logic_vector(15 DOWNTO 0); -- 7 segment data
  dio       : IN  std_logic_vector(15 DOWNTO 0); -- Diodes data
  btns      : OUT std_logic_vector(14 DOWNTO 0); -- 15 Buttons
  swtch     : OUT std_logic_vector(7 DOWNTO 0);  -- 8 Switches
  lcd_on    : IN  std_logic;
  lcd       : IN  std_logic_vector(63 downto 0)    -- 16 x 4bits to display on LCD
       );
END dio2_interface;

ARCHITECTURE rtl OF dio2_interface IS

TYPE t_asc_table IS array (0 to 15) of std_logic_vector(7 downto 0);

CONSTANT ascii_table : t_asc_table :=
  (X"30", X"31", X"32", X"33", X"34", X"35", X"36", X"37", X"38", X"39",
   X"41", X"42", X"43", X"44", X"45", X"46");
 -- ASCII table for converting numbers to ASCII codes

  SIGNAL res_int : std_logic;                          -- Internal reset
  SIGNAL startup : std_logic_vector(4 DOWNTO 0);       -- Startup reset generator
  SIGNAL prescaler : std_logic_vector(24 DOWNTO 0);    -- Source of all internal clock signals
  SIGNAL lcd_ready : std_logic;                        -- Various LCD signals ----->
  SIGNAL lcd_addr, reg_addr : std_logic_vector(6 DOWNTO 0);
  SIGNAL lcd_data, reg_data : std_logic_vector(7 DOWNTO 0);
  SIGNAL lcd_char : std_logic;
    
  
  
BEGIN

---- constant subtitutions
res<= res_int;
dio2clk <= prescaler(10); -- 50 MHz/2^11 = 24 KHz
clocks<=prescaler;
addr(6 DOWNTO 3) <= (OTHERS => '0');

-- Prescaler process
clk_Counter:PROCESS(clk, res_int)
BEGIN
IF res_int = '1' THEN
   prescaler <= (OTHERS => '0');
ELSIF clk'event AND clk = '1' THEN
   prescaler <= prescaler + 1;
END IF;
END PROCESS;


-- Counts several clock cycles before releasing global reset signal.
starter:PROCESS(clk, reset)
BEGIN
  IF reset='1' THEN
    startup<=(OTHERS=>'0');
    res_int<='1';
  ELSIF clk'event AND clk = '1' THEN
    IF startup = 16 THEN
      res_int <= '0';
      startup <= startup;
    ELSE
      res_int <= '1';
      startup <= startup + '1';
    END IF;
  END IF;
END PROCESS;

lcd_addr<="0000000";

lcd_data_mux: PROCESS(prescaler, lcd)
variable number : natural;
variable sel    : std_logic_vector(3 downto 0);
variable lcd_sel: std_logic_vector(3 downto 0);
BEGIN
  sel:=not(prescaler(15 downto 12));
  CASE sel IS
    WHEN "0000" => lcd_sel:=lcd(3 downto 0);
    WHEN "0001" => lcd_sel:=lcd(7 downto 4);
    WHEN "0010" => lcd_sel:=lcd(11 downto 8);
    WHEN "0011" => lcd_sel:=lcd(15 downto 12);
    WHEN "0100" => lcd_sel:=lcd(19 downto 16);
    WHEN "0101" => lcd_sel:=lcd(23 downto 20);
    WHEN "0110" => lcd_sel:=lcd(27 downto 24);
    WHEN "0111" => lcd_sel:=lcd(31 downto 28);
    WHEN "1000" => lcd_sel:=lcd(35 downto 32);
    WHEN "1001" => lcd_sel:=lcd(39 downto 36);
    WHEN "1010" => lcd_sel:=lcd(43 downto 40);
    WHEN "1011" => lcd_sel:=lcd(47 downto 44);
    WHEN "1100" => lcd_sel:=lcd(51 downto 48);
    WHEN "1101" => lcd_sel:=lcd(55 downto 52);
    WHEN "1110" => lcd_sel:=lcd(59 downto 56);
    WHEN "1111" => lcd_sel:=lcd(63 downto 60);
    WHEN OTHERS => lcd_sel:=(OTHERS=>'-');
  END CASE;
  number:=conv_integer(lcd_sel);
  lcd_data<=ascii_table(number);
END PROCESS;

--CPLD/LCD control process
--Manages writing and reading data from CPLD and LCD included on the DIO2 board.
cpld_control:PROCESS(clk, res_int)
BEGIN
  IF res_int = '1' THEN
    oe <= '0';
    we <= '0';
    cs <= '0';
    lcd_rs <= '0';
    lcd_rw <= '0';
    lcd_e <= '0';
    lcd_ready <= '0';
    addr(2 DOWNTO 0) <= (OTHERS => '0');
    data <= (OTHERS => 'Z');
    btns <= (OTHERS => '0');
    swtch <= (OTHERS => '0');
  ELSIF clk'event AND clk = '1' THEN
    IF lcd_ready = '0' THEN
      IF prescaler = 1003 THEN
        lcd_e <= '1';
        data <= "00111000";
      ELSIF prescaler = 2000 THEN
        lcd_e <= '0';
      ELSIF prescaler = 5000 THEN
        lcd_e <= '1';
        data <= "00001100";
      ELSIF prescaler = 5050 THEN
        lcd_e <= '0';
      ELSIF prescaler = 9000 THEN
        lcd_e <= '1';
        data <= "00000001";
      ELSIF prescaler = 9050 THEN
        lcd_e <= '0';
      ELSIF prescaler > 1048576 THEN
        lcd_e <= '0';
        lcd_ready <= '1';
        data <= (OTHERS => 'Z');
      ELSE
        lcd_ready <= '0';
      END IF;    
    ELSIF lcd_ready = '1' THEN
      IF prescaler(16 DOWNTO 11) = 0 THEN
        cs <= '1';
        we <= '1';
        oe <= '0';
        addr(2 DOWNTO 0) <= "100"; -- diodes part1
        data (7)<= dio(8);
        data (6)<= dio(9);
        data (5)<= dio(10);
        data (4)<= dio(11);
        data (3)<= dio(12);
        data (2)<= dio(13);
        data (1)<= dio(14);
        data (0)<= dio(15);
    ELSIF prescaler(16 DOWNTO 11) = 1 THEN
        we <= '0';
        oe <= '0';
        data <= (OTHERS => 'Z'); 
    ELSIF prescaler(16 DOWNTO 11) = 2 THEN
        we <= '1';
        oe <= '0';
        addr(2 DOWNTO 0) <= "101";
        data (7)<= dio(0);
        data (6)<= dio(1);
        data (5)<= dio(2);
        data (4)<= dio(3);
        data (3)<= dio(4);
        data (2)<= dio(5);
        data (1)<= dio(6);
        data (0)<= dio(7); -- diodes part2
    ELSIF prescaler(16 DOWNTO 11) = 3 THEN
        we <= '0';
        oe <= '0';
        data <= (OTHERS => 'Z'); 
    ELSIF prescaler(16 DOWNTO 11) = 4 THEN
        we <= '0';
        oe <= '1';
        addr(2 DOWNTO 0) <= "000";
        data <= (OTHERS => 'Z');
    ELSIF prescaler(16 DOWNTO 11) = 5 THEN
        we <= '0';
        oe <= '1';
        addr(2 DOWNTO 0) <= "000";
        data <= (OTHERS => 'Z');
        btns(7 DOWNTO 0) <= data; -- reading buttons part 1
    ELSIF prescaler(16 DOWNTO 11) = 5 THEN
        we <= '0';
        oe <= '0';
        data <= (OTHERS => 'Z');
    ELSIF prescaler(16 DOWNTO 11) = 6 THEN
        we <= '0';
        oe <= '1';
        addr(2 DOWNTO 0) <= "001";
        data <= (OTHERS => 'Z');
    ELSIF prescaler(16 DOWNTO 11) = 7 THEN
        we <= '0';
        oe <= '1';
        addr(2 DOWNTO 0) <= "001";
        data <= (OTHERS => 'Z');
        btns(14 DOWNTO 8) <= data(6 downto 0); -- reading buttons part 2
    ELSIF prescaler(16 DOWNTO 11) = 8 THEN
        we <= '0';
        oe <= '0';
        data <= (OTHERS => 'Z');
    ELSIF prescaler(16 DOWNTO 11) = 9 THEN
        we <= '0';
        oe <= '1';
        addr(2 DOWNTO 0) <= "010";
        data <= (OTHERS => 'Z');
    ELSIF prescaler(16 DOWNTO 11) = 10 THEN
        we <= '0';
        oe <= '1';
        addr(2 DOWNTO 0) <= "010";
        data <= (OTHERS => 'Z');
        swtch(7) <= data(0);
        swtch(6) <= data(1);
        swtch(5) <= data(2);
        swtch(4) <= data(3);
        swtch(3) <= data(4);
        swtch(2) <= data(5);
        swtch(1) <= data(6);
        swtch(0) <= data(7);
    ELSIF prescaler(16 DOWNTO 11) = 11 THEN
        we <= '0';
        oe <= '0';
        data <= (OTHERS => 'Z');
    ELSIF prescaler(16 DOWNTO 11) = 12 THEN
        we <= '1';
        oe <= '0';
        addr(2 DOWNTO 0) <= "110";
        data (7 downto 4) <= sseg(11 DOWNTO 8); -- 7S digits 3 & 4
        data (3 downto 0) <= sseg(15 downto 12);
    ELSIF prescaler(16 DOWNTO 11) = 13 THEN
        we <= '0';
        oe <= '0';
        data <= (OTHERS => 'Z');
    ELSIF prescaler(16 DOWNTO 11) = 14 THEN
        we <= '1';
        oe <= '0';
        addr(2 DOWNTO 0) <= "111";
        data (7 downto 4)<= sseg(3 DOWNTO 0); -- 7S digits 1 & 2
        data (3 downto 0)<= sseg(7 downto 4);
    ELSIF prescaler(16 DOWNTO 11) = 15 THEN
        cs <= '0';
        we <= '0';
        oe <= '0';
        data <= (OTHERS => 'Z');
--- LCD ----
    ELSIF prescaler(16 DOWNTO 11) = 16 AND lcd_on = '1' THEN
        cs <= '0';
        lcd_e <= '0';
        lcd_rs <= '0';
        lcd_rw <= '0';
        data <= (OTHERS => 'Z');
    ELSIF prescaler(16 DOWNTO 11) = 17 AND lcd_on = '1' THEN
        cs <= '0';
        lcd_e <= '1';
        lcd_rs <= '0';
        lcd_rw <= '0';
        data <= '1' & lcd_addr; -- set DDRAM address
    ELSIF prescaler(16 DOWNTO 11) = 18 AND lcd_on = '1' THEN
        cs <= '0';
        lcd_e <= '0';
        lcd_rs <= '0';
        lcd_rw <= '0';
    ELSIF (prescaler(16)='1') AND (prescaler(11)='0') AND lcd_on = '1' THEN
        cs <= '0';
        lcd_e <= '1';
        lcd_rs <= '1';
        lcd_rw <= '0';
        data <= lcd_data;
   ELSIF (prescaler(16)='1') AND (prescaler(11)='1') AND lcd_on = '1' THEN
        cs <= '0';
        lcd_e <= '0';
        lcd_rs <= '1';
        lcd_rw <= '0';
        data <= lcd_data;
    ELSIF prescaler(16 DOWNTO 11) = 20 AND lcd_on = '1' THEN
        cs <= '0';
        lcd_e <= '0';
      ELSE  cs <= '0';
            we <= '0';
            oe <= '0';
            lcd_e <= '0';
            lcd_rs <='0';
            addr(2 DOWNTO 0) <= (OTHERS => '0');
            data <= (OTHERS => 'Z');
      END IF;
    END IF;
  END IF;
END PROCESS;


END rtl;

